# -*- coding: utf-8 -*-

"""
Copyright(C) 2007-2008 INL
Written by Romain Bignon <romain AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

$Id: datasource.py 17754 2009-02-03 09:09:00Z romain $
"""

from cStringIO import StringIO
import datetime
import socket
from twisted.enterprise.adbapi import safe
from twisted.internet import defer
from IPy import IP

class NuCentralError(Exception):

    def __init__(self, s, *args):
        if args:
            Exception.__init__(self, s % args)
        else:
            Exception.__init__(self, s)

# There isn't (yet) any translation from backend
def tr(s):
    return s

class DataSource:

    def __init__(self, database):

        self.database = database
        self.args = {}
        self.filters = {}

    def getFiltersList(self):
        return ()

    def ip2str(self, ip):
        """ Get an IP integer and return a formated IP string
            @param ip [integer]
            @return [string]
        """

        if self.database.dbtype == 'pgsql':
            return ip

        if self.database.ip_type == 6:
            i = 0
            for j in xrange(len(ip)):
                i |= ord(ip[j]) << (8 * (len(ip) - j - 1))

            ipret = IP(i, 6)
            if ipret.iptype() == 'IPV4MAP':
                return '::ffff:' + IP('0x'+ipret.strFullsize().lstrip('0000:0000:0000:0000:0000:ffff:').replace(':', '')).strCompressed()
            else:
                return ipret.strCompressed()
        else:
            try:
                return IP(int(ip), 4).strCompressed()
            except ValueError:
                raise NuCentralError(tr('Incorrect IP format. Please change IP type value (4 -> 6) in NuLog configuration'))

    def str2ip(self, string):
        """ Get a string to return an IP integer
            @param string [string]  IP string
            @return [IP]  an IP object
        """
        if self.database.ip_type == 6:
            if IP(string).iptype() == 'IPV4COMP':
                string = '::ffff:' + string.lstrip(':')
        return IP(string)

    def _sql_query(self, functioname, *args, **kwargs):
        """ Send a query to SGDB and receive answer in _print_result() method.
            @param query [string] SQL formated query

            @return [Deferred] This is a deferred object. If this method is called by a
                               SOAP client, return this Deferred objet to tell SOAP to
                               wait callback result to send it to client.
        """

        start_time = None
        end_time = None
        query = StringIO()

        # Actual 'ulog' table is from the first packet (0) and now.
        self.database.archives[self.database.ulog] = (self.database.archives[self.database.ulog][0], datetime.datetime.today())

        if self.args.has_key('start_time'):
            start_time = datetime.datetime.fromtimestamp(int(self.args['start_time']))
        else:
            start_time = self.database.archives[self.database.ulog][0]

        if self.args.has_key('end_time'):
            end_time = datetime.datetime.fromtimestamp(int(self.args['end_time']))
        else:
            end_time = self.database.archives[self.database.ulog][1]

        assert start_time
        assert end_time

        display = True
        if kwargs.has_key('display') and not kwargs['display']:
            display = False
            kwargs.pop('display')

        for table, date in self.database.archives.items():

            if date[1] and date[0] and start_time < date[1] and end_time > date[0]:

                request = self.database.createRequest(table)
                if query.getvalue():
                    query.write(' UNION ')

                query.write(getattr(request, functioname)(*args, **kwargs))

        if not query.getvalue():
            # There isn't any query, so we exit
            return defer.succeed(self)

        if display:
            if self.args.has_key('sortby') and self.args.has_key('sort'):
                query.write(' ORDER BY %s %s' % (self.args['sortby'], self.args['sort']))
            if self.args.has_key('start') and self.args['start'] >= 0 and self.args.has_key('limit') and self.args['limit'] >= 0:
                query.write(' LIMIT %s OFFSET %d' % (self.args['limit'], self.args['start']))

        return self.database.query(query.getvalue())

    def _arg_int(self, args, argname):
        """ Check if argname is in args, and if it is an integer
            @param args [dict] Args where we look for argname
            @param argname [string] Argument we want to check
            @return NOTHING. Value is changed in self.args
        """

        if not args.has_key(argname):
            if not self.args.has_key(argname):
                raise NuCentralError(tr("Missing argument '%s'"), argname)
            else:
                return

        value = args[argname]
        try:
            self.args[argname] = int(value)
        except:
            raise NuCentralError(tr('%s must be an integer'), argname)

    def _arg_bool(self, args, argname):

        if not args.has_key(argname):
            if not self.args.has_key(argname):
                raise NuCentralError(tr("Missing argument '%s'"), argname)
            else:
                return

        value = args[argname]
        if value == 1 or value is True or value == '1' or value == 'true' or value == 'True' or value == 'yes' or value == 'y':
            self.args[argname] = True
        else:
            self.args[argname] = False

    def _arg_in(self, args, argname, lst):
        """ Look for argname in args, and check if it is in lst.
            @param args [dict] Args
            @param argname [string] Argument name
            @param lst [list] Value may be in this list.
            @return NOTHING. Value is changed in self.args
        """

        if not args.has_key(argname):
            return

        if args[argname] in lst:
            self.args[argname] = args[argname]

    def _arg_where(self, args, where, lst):
        """ Make a SQL WHERE string with args and a definition of args (in lst).
            @param args [dict] Args
            @param where [StringIO (!)] We will put here WHERE string. Note that if where has already
                                        a value, we will add our string in it (with " AND ..")
            @param lst [dict] This dict contains an association between arg and function to call
            @return NOTHING.

            Note: help function must have a prototype like :
                     def help_function(args[DICT], key[STRING], value[STRING])
                  You can use one of helper function below.
        """

        for w in lst:
            if args.has_key(w):
                if args[w] is None or args[w] == '':
                    continue # We don't take empty strings

                if not where.getvalue():
                    entry = "WHERE "
                else:
                    entry = " AND "

                value = args[w]
                if isinstance(value, unicode):
                    value = value.encode("utf8")

                # There isn't any function to call
                if lst[w] is None:
                    entry += "%s = " % w

                    if isinstance(value, int) or isinstance(value, long) or isinstance(value, str) and value.isdigit():
                        entry += str(value)
                    else:
                        entry += "'%s'" % safe(value)
                else:
                    try: # We try to call function, and if there is an exception raised, we show backtrace and raise it.
                        ret = lst[w](args, w, safe(str(value)))
                        if not ret:
                            continue

                        entry += ret
                    except Exception, e:
                        raise NuCentralError(tr('An error has occured: %s'), str(e))

                self.args[w] = args[w]
                self.filters[w] = args[w]

                where.write(entry)

    def __arg_where_priv_ip(self, args, key, value, function):
        """ Never call this function outside of _arg_where_ip* functions. """

        try:
            ips = socket.gethostbyname_ex(value)[2]
        except socket.gaierror:
            try:
                ip = IP(value)
                if ip.version() > self.database.ip_type:
                    raise ValueError('IP address must be an IPv%d' % self.database.ip_type)
            except ValueError, e:
                raise NuCentralError(tr('Please give a correct hostname or IP: %s') % e)
            ips = [value]

        s = ''
        for ip in ips:
            if s:
                s += ' OR '

            ip = self.str2ip(ip)

            s += '(%s)' % function(args, key, ip)

        return '(%s)' % s

    def _arg_where_ip(self, args, key, value):
        # Hack: when we are filtering by ip_daddr or ip_saddr, make it affect an ip_addr column as well
        if "ip_addr" in self.columns:
            return '(%s OR %s)' % (self._arg_where_priv_ip(args, 'ip_saddr', value, self.database.whereIP),
                                   self._arg_where_priv_ip(args, 'ip_daddr', value, self.database.whereIP))
        return self.__arg_where_priv_ip(args, key, value, self.database.whereIP)

    def _arg_where_REVERSEip(self, args, key, value):
        return self.__arg_where_priv_ip(args, key, value, self.database.whereIP_REVERSE)

    def _arg_where_ip_both(self, args, key, value):

        if not args.has_key('ip_from') or not args['ip_from'] in ('s', 'd'):
            return '(%s OR %s)' % (self._arg_where_ip(args, 'ip_saddr', value),
                                   self._arg_where_ip(args, 'ip_daddr', value))
        else:
            return self._arg_where_ip(args, ('ip_%saddr' % args['ip_from']), value)

    def _arg_where_port(self, args, key, value):

        if not isinstance(value, int) and not isinstance(value, long) and (not isinstance(value, str) or not value.isdigit()):
            raise NuCentralError(tr('Please specify an integer value'))

        protos = ('tcp', 'udp')
        if not args.has_key('proto') or not args['proto'] in protos:
            return '(tcp_%s = %s OR udp_%s = %s)' % (key, value, key, value)
        else:
            return '%s_%s = %s' % (args['proto'], key, value)

    def _arg_where_int(self, args, key, value):

        if not isinstance(value, int) and not isinstance(value, long) and (not isinstance(value, str) or not value.isdigit()):
            raise NuCentralError(tr('Please specify an integer value'))

        return '%s = %s' % (key, value)

    def _arg_where_proto(self, args, key, value):

        if value in ('tcp', 'udp'):
            return '%s_sport IS NOT NULL' % value
        elif value == 'icmp':
            return 'tcp_sport IS NULL AND udp_sport IS NULL'
        else:
            raise NuCentralError(tr('Protocol must be tcp or udp (is %s)'), value)

    def _arg_where_state(self, args, key, value):

        try:
            value = int(value)

            if value == -1:
                # -1 = ALL
                return ''

            if value < 0 or value > 4:
                raise NuCentralError(tr('State must be an integer between 0 and 4.'))

            if value == 4:
                return '(%s = 1 OR %s = 2)' % (key, key)
            else:
                return '%s = %s' % (key, value)
        except:
            raise NuCentralError(tr('Please specify an integer value'))

    def _arg_where_begin_time(self, args, key, value):

        if not isinstance(value, int) and not isinstance(value, long) and (not isinstance(value, str) or not value.isdigit()):
            raise NuCentralError(tr("Please specify an integer value"))

        return 'oob_time_sec >= %s' % value

    def _arg_where_end_time(self, args, key, value):

        if not isinstance(value, int) and not isinstance(value, long) and (not isinstance(value, str) or not value.isdigit()):
            raise NuCentralError(tr("Please specify an integer value"))

        return 'oob_time_sec <= %s' % value

    def _arg_where_like(self, args, key, value):

        return '%s LIKE \'%%%s%%\'' % (key, value)


