# -*- coding: utf-8 -*-

"""
Copyright(C) 2007-2008 INL
Written by Romain Bignon <romain AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

$Id: nulog.py 17058 2008-12-01 10:24:23Z romain $
"""

import table
from database import DataBaseFactory, DatabaseError

try:
    from nucentral.core import cron
except:
    from wrapper.core import cron

import info
from datetime import datetime

from twisted.internet.defer import DeferredList
import sys, traceback, os
from ConfigParser import SafeConfigParser

from logging import error

update_archives_period = 86400 # 60*60*24 = 1j
etcpath = '/etc/nulog/'
mypath = os.path.dirname(__file__) + os.sep
core_conf = "core.conf"
core_defconf = "default.core.conf"
update_conntrack_period= 60
update_conntrack_newtimeout = 120

def tr(msg): return msg

def getBacktrace(empty="Empty backtrace."):
    """
    Try to get backtrace as string.
    Returns "Error while trying to get backtrace" on failure.
    """
    try:
        info = sys.exc_info()
        trace = traceback.format_exception(*info)
        sys.exc_clear()
        if trace[0] != "None\n":
            return "".join(trace)
    except:
        # No i18n here (imagine if i18n function calls error...)
        return "Error while trying to get backtrace"
    return empty

class NulogCore:

    tables = {
                'TCPTable':             table.TCPTable,
                'UDPTable':             table.UDPTable,
                'IPsrcTable':           table.IPsrcTable,
                'IPdstTable':           table.IPdstTable,
                'UserTable':            table.UserTable,
                'PacketTable':          table.PacketTable,
                'ConUserTable':         table.ConUserTable,
                'AppTable':             table.AppTable,
                'PacketInfo':           info.PacketInfo,
                'BadHosts':             table.BadHosts,
                'BadUsers':             table.BadUsers,
            }

    default_conf = {
             'host':      'localhost',
             'db':        'ulog',
             'user':      'nulog',
             'password':  'nulog',
             'dbtype':    'mysql',
             'type':      'triggers',
             'ip':        '4',
             'table':     'ulog',
             'conntrack': 'conntrack_ulog',
             'maxrotate': '52',
        }

    def __init__(self, parser):

        self.conf = self.default_conf
        for key in self.conf.keys():
            self.conf[key] = parser.get('DB', key)

        self.database = None
        self.tiny_does_not_hide_timestamp = int(parser.get('Misc', 'tiny_does_not_hide_timestamp'))
        self.archives_cron = None

        try:
            self.start_dbpool()
        except DatabaseError, err:
            # We don't stop nucentral launch because this isn't a problem,
            # user will correct it with nulog-qt and the setConfig() service.
            self.writeError(err, "start_dbpool() error")

    def writeError(self, err, msg):

        error('%s: %s' % (msg, err))

    def start_dbpool(self):

        if not self.database or self.database.dbtype != self.conf['dbtype']:
            try:
                self.database = DataBaseFactory().create(self.conf, self)
            except DatabaseError, err:
                self.database = None
                self.writeError(err, "Unable to connect to the database")
                raise err
        else:
            self.database.conf = self.conf

        self.database.start()

        self.database._type = self.conf['type']
        self.database.ip_type = int(self.conf['ip'])
        self.database.ulog = self.conf['table']
        self.database.archives = {}
        self.database.tiny_does_not_hide_timestamp = self.tiny_does_not_hide_timestamp

        if self.archives_cron:
            self.archives_cron.stop()
        self.archives_cron = cron.scheduleRepeat(update_archives_period, self.update_archives)

    def update_archives(self, result=None, nb_table=0):
        """
            This function is used to update cache of archives list.
            We store the first and last packet timestamp, to know
            in which time interval this archive table is used.

            Call it without any param to launch the process, and it
            will call itself as callback to iterate all archive table.

            It begins with ulog_1 and increment to the latest existant table.
        """

        # There is not any database connection.
        if not self.database or not self.database.dbpool:
            return

        def eb(result):
            # We doesn't care about an error..
            pass

        if result:
            # If result != None it is because this is a callback,
            # so we can store it.

            try:
                assert result[0][0]
                assert result[0][1]

                self.database.archives['%s_%d' % (self.database.ulog, nb_table)] = result[0]
                if self.database.archives[self.database.ulog][0] == datetime.min:
                    self.database.archives[self.database.ulog] = (result[0][1], datetime.today())
            except:
                pass

        else:
            self.database.archives[self.database.ulog] = (datetime.min, datetime.today())

        table = '%s_%d' % (self.database.ulog, nb_table+1)

        return self.database.query("""SELECT MIN(oob_time_sec), MAX(oob_time_sec) FROM %s"""
                                    % table).addCallback(self.update_archives, nb_table+1).addErrback(eb)

    def check_dbpool_running(self):

        if not self.database or not self.database.dbpool:
            self.start_dbpool()

    ##########
    # SERVICES

    def service_count_entries(self, what):
        """ Use a keyword to tell what request you want to do on database.
            Goal of this function is only to count entries with filters
        """

        self.check_dbpool_running()

        def cb(result):
            return result[0]

        request = self.database.createRequest(self.database.ulog)

        if hasattr(request, 'count_%s' % what):
            query = getattr(request, 'count_%s' % what)()

            return self.database.query(query).addCallback(cb)
        else:
            raise Exception("Unable to find this value: %s" % what)

    def service_table(self, name, args):
        """ This is the 'table' distant access function.
            @param name [string] function name;
            @param args [dictionnary]
        """

        self.check_dbpool_running()

        p = self.tables.get(name) # We get class type

        if p is None:
            raise Exception(tr('Table not found:'), name)

        try:
            p = p(self.database) # We instance class

            defer = p(**args) # <=> p.__call__(..)

            return defer

        except Exception, err:
            self.writeError(err, "table() error")
            raise err

core = None

def getComponentName():
    """
        Function called by NuCentral (or wrapper) to get my name
        I also create database connection, because I know this function
        is called at program launch.
    """

    parser = SafeConfigParser()

    parser.read([etcpath + core_defconf, mypath + core_defconf, etcpath + core_conf, mypath + core_conf])

    global core
    core = NulogCore(parser)

    return 'nulog-core'

def getServiceList():
    """ Called by NuCentral, I return my services """

    services = dict()
    services['table'] = core.service_table
    services['count'] = core.service_count_entries
    return services

