# -*- coding: utf-8 -*-

"""
Copyright(C) 2007-2008 INL
Written by Romain Bignon <romain AT inl.fr>
           Pierre Chifflier <chifflier AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

$Id$
"""

from twisted.enterprise.adbapi import safe
import re

#############################################
#                                           #
#             Request Object                #
#                                           #
#############################################
class Request:
    """ This class is used to abstract SQL requests.
        Use the same interface on a child class to make other requests..

        nulog.py decides what class is created.
    """

    def __init__(self, database, tablename, ip_type):
        """ Create a database request
            @param database [DataBase] database object
            @param tablename [string] ulog table name (really usefull?)
            @param ip_type [integer] 4 or 6.

            TODO: for now it isn't really useful to give the "tablename"
                  argument because it is database.ulog, but... it can
                  become useful soon.
        """

        self.database = database
        self.ulog = safe(tablename)

        # This is usefull to know if we work with an archive table, and
        # which prefix use, for example for usersstats table.
        m = re.match("%s(.*)" % self.database.ulog, self.ulog)
        self.sufix = m and m.groups()[0] or ''

        self.ip_type = int(ip_type)

    def select_packets(self, where):
        """ Method used to select all packets which matches "where" clause. """

        table = self.ulog

        return """SELECT id as packet_id, username, user_id, ip_saddr, ip_daddr,
                    CASE WHEN tcp_sport IS NOT NULL THEN 'tcp'
                         WHEN udp_sport IS NOT NULL THEN 'udp'
                         ELSE 'icmp'
                    END as proto,
                    COALESCE(tcp_sport, udp_sport) as sport,
                    COALESCE(tcp_dport, udp_dport) as dport,
                    oob_time_sec, oob_prefix, state
                    FROM %s
                    %s""" % (table, where)

    def count_packets(self, where):

        return """SELECT COUNT(*) FROM %s %s """ % (self.ulog, where)

    def get_packet(self, _id):
        """ Get a specific packet with his id """

        return """SELECT id as packet_id, username, user_id, oob_time_sec, oob_time_usec, oob_in, oob_out,
                        oob_prefix, oob_mark, ip_saddr, ip_daddr, ip_tos, ip_ttl, ip_totlen,
                        ip_ihl, ip_csum, ip_id,
                        CASE WHEN tcp_sport IS NOT NULL THEN 'tcp'
                            WHEN udp_sport IS NOT NULL THEN 'udp'
                            ELSE 'icmp'
                        END as protocol,
                        COALESCE(tcp_sport, udp_sport) as sport,
                        COALESCE(tcp_dport, udp_dport) as dport,
                        tcp_seq, tcp_ackseq, tcp_window, tcp_urg, tcp_urgp, tcp_ack, tcp_psh, tcp_rst, tcp_syn, tcp_fin,
                        udp_len, icmp_type, icmp_code, client_os, client_app, bytes_in, bytes_out,
                        packets_in, packets_out, raw_mac, state
                    FROM %s
                    WHERE id = %s
                """ % (self.ulog, _id)

    def select_conusers(self, where):
        """ Get all connected users """

        return """SELECT username, user_id, ip_saddr, os_sysname, start_time, end_time
                    FROM users%s
                    %s""" % (self.sufix, where)

    def select_ports(self, proto, where):
        """ List all ports """

        if where:
            where += ' AND state = 0 AND %s_dport IS NOT NULL' % proto
        else:
            where = 'WHERE state = 0 AND %s_dport IS NOT NULL' % proto

        return """SELECT %s_dport, COUNT(*) AS packets, MIN(oob_time_sec) AS start_time, MAX(oob_time_sec) AS end_time
                    FROM %s
                    %s
                    GROUP BY %s_dport
                """ % (proto, self.ulog, where, proto)

    def count_drop_port(self, proto, where):
        """ Count number of dropped ports on this protocol
            @param proto [string] 'tcp' or 'udp'
        """

        if where:
            where += ' AND state = 0 AND %s_dport IS NOT NULL' % proto
        else:
            where = 'WHERE state = 0 AND %s_dport IS NOT NULL' % proto

        return """SELECT COUNT(*) FROM %s %s""" % (self.ulog, where)

    def select_ip(self, direction, where):
        """ List all ips which matches filters """

        if where:
            where += ' AND state = 0 AND ip_%saddr IS NOT NULL' % direction
        else:
            where = 'WHERE state = 0 AND ip_%saddr IS NOT NULL' % direction

        return """SELECT ip_%saddr, COUNT(*) AS packets, MIN(oob_time_sec) AS start_time, MAX(oob_time_sec) AS end_time
                    FROM %s
                    %s
                    GROUP BY ip_%saddr
                """ % (direction, self.ulog, where, direction)

    def count_drop_ip(self, direction, where):
        """ Count all dropped packets """

        if where:
            where += ' AND state = 0 AND ip_%saddr IS NOT NULL' % direction
        else:
            where = 'WHERE state = 0 AND ip_%saddr IS NOT NULL' % direction

        return """SELECT COUNT(*) FROM %s %s""" % (self.ulog, where)

    def select_apps(self, where):

        if where:
            where += ' AND state = 0 AND client_app IS NOT NULL'
        else:
            where = 'WHERE state = 0 AND client_app IS NOT NULL'

        return """SELECT client_app, COUNT(*) AS packets, MIN(oob_time_sec) AS start_time, MAX(oob_time_sec) AS end_time
                    FROM %s
                    %s
                    GROUP BY client_app
                """ % (self.ulog, where)

    def count_drop_apps(self, where):
        """ Count all dropped packets """

        if where:
            where += ' AND state = 0 AND client_app IS NOT NULL'
        else:
            where = 'WHERE state = 0 AND client_app IS NOT NULL'

        return """SELECT COUNT(*) FROM %s %s""" % (self.ulog, where)

    def select_user(self, where):
        """ List all users who have dropped packets """

        if where:
            where += ' AND state = 0 AND user_id IS NOT NULL'
        else:
            where = 'WHERE state = 0 AND user_id IS NOT NULL'

        return """SELECT username, user_id, COUNT(*) AS packets,
                         MIN(oob_time_sec) AS start_time, MAX(oob_time_sec) AS end_time
                    FROM %s
                    %s
                    GROUP BY user_id,username
                """ % (self.ulog, where)

    def count_drop_user(self, where):
        """ Count all users packets which are dropped """

        if where:
            where += ' AND state = 0 AND user_id IS NOT NULL'
        else:
            where = 'WHERE state = 0 AND user_id IS NOT NULL'

        return """SELECT COUNT(*) FROM %s %s""" % (self.ulog, where)

    def select_badhosts(self):

        return """SELECT ip_saddr, count(*)/300 as RATE
                  FROM %s
                  WHERE (oob_time_sec > NOW() - INTERVAL 5 MINUTE) AND state = 0
                  GROUP BY ip_saddr""" % self.ulog

    def select_badusers(self):

        return """SELECT user_id, username, count(*)/300 as RATE
                  FROM %s
                  WHERE oob_time_sec > NOW()- INTERVAL 5 MINUTE AND username NOT LIKE \"\" AND state=0
                  GROUP BY user_id""" % self.ulog

    def count_average(self, minutes):

        return """SELECT COUNT(*)/%d FROM %s WHERE oob_time_sec >= UNIX_TIMESTAMP(Now()) - %d AND state = 0""" % (minutes*60, self.ulog, minutes*60)

    def count_average1(self): return self.count_average(1)
    def count_average5(self): return self.count_average(5)
    def count_average15(self): return self.count_average(15)

    def count_conusers(self):
        return  'SELECT COUNT(*) FROM users WHERE end_time IS NULL'

    def count_packets(self):
        return 'SELECT COUNT(*) FROM %s' % self.ulog

#############################################
#                                           #
#          Ulogd2Request Object             #
#                                           #
#############################################

class Ulogd2Request(Request):

    def get_packet(self, _id):
        """ Get a specific packet with his id """

        return """SELECT id as packet_id, username, user_id, oob_time_sec, oob_time_usec, oob_in, oob_out,
                        oob_prefix, oob_mark, ip_saddr, ip_daddr, ip_tos, ip_ttl, ip_totlen,
                        ip_ihl, ip_csum, ip_id,
                        CASE WHEN tcp_sport IS NOT NULL THEN 'tcp'
                            WHEN udp_sport IS NOT NULL THEN 'udp'
                            ELSE 'icmp'
                        END as protocol,
                        COALESCE(tcp_sport, udp_sport) as sport,
                        COALESCE(tcp_dport, udp_dport) as dport,
                        tcp_seq, tcp_ackseq, tcp_window, tcp_urg, tcp_urgp, tcp_ack, tcp_psh, tcp_rst, tcp_syn, tcp_fin,
                        udp_len, icmp_type, icmp_code, client_os, client_app, raw_mac
                    FROM %s
                    WHERE id = %s
                """ % (self.ulog, _id)


#############################################
#                                           #
#          TriggerRequest Object            #
#                                           #
#############################################

class TriggerRequest(Request):
    """ This class, child of DataBase, is used to make SQL requests on help tables 'offenders',
        'tcp_ports, 'udp_ports' and 'usersstats'.
    """

    def select_ports(self, proto, where):

        if where or self.sufix:
            # We can only show all dropped packets, without any filter.
            return Request.select_ports(self, proto, where)

        return """SELECT %s_dport, count AS packets, first_time AS start_time, last_time AS end_time
                    FROM %s_ports """ % (proto, proto)

    def count_drop_port(self, proto, where):

        if where or self.sufix:
            return Request.count_drop_port(self, proto, where)

        return """SELECT SUM(count) FROM %s_ports""" % proto

    def select_ip(self, direction, where):

        if where or direction != 's' or self.sufix:
            return Request.select_ip(self, direction, where)

        return """SELECT ip_addr AS ip_saddr, count AS packets, first_time AS start_time, last_time AS end_time
                    FROM offenders
               """

    def count_drop_ip(self, direction, where):

        if where or direction != 's' or self.sufix:
            return Request.count_drop_ip(self, direction, where)

        return """SELECT SUM(count) FROM offenders"""

    def select_user(self, where):

        if where or self.sufix:
            return Request.select_user(self, where)

        return """SELECT username, user_id, bad_conns AS packets, first_time AS start_time, last_time AS end_time
                    FROM usersstats
                    WHERE bad_conns > 0
                """

    def count_drop_user(self, where):

        if where or self.sufix:
            return Request.count_drop_user(self, where)

        return """SELECT SUM(bad_conns) FROM usersstats"""

#############################################
#                                           #
#             PostgreSQL Object             #
#                                           #
#############################################

class PGRequest(Request):

    def select_badhosts(self):

        return """SELECT ip_saddr, count(*)/300 as RATE
                  FROM %s
                  WHERE (oob_time_sec > EXTRACT(epoch FROM NOW() - INTERVAL '5 MINUTE')) AND state = 0
                  GROUP BY ip_saddr""" % self.ulog

    def count_average(self, minutes):

        return """SELECT COUNT(*)/%d
                  FROM %s
                  WHERE oob_time_sec > EXTRACT(epoch FROM NOW()- INTERVAL \'%d MINUTE\')
                        AND state = 0""" % (minutes*60, self.ulog, minutes*60)

    def count_conusers(self):
        return  'SELECT COUNT(*) FROM users WHERE end_time IS NULL'

class PGUlogd2Request(PGRequest, Ulogd2Request):

    def get_packet(self, _id):
        return Ulogd2Request.get_packet(self, _id)

class PGTriggerRequest(PGRequest):#, TriggerRequest):

    # TODO: for now, triggers don't work with pgsql.
    # PGRequest methods will have priority on overriding.
    pass
