# -*- coding: utf-8 -*-

"""
Copyright(C) 2007-2008 INL
Written by Romain Bignon <romain AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

$Id: tablebase.py 16911 2008-11-20 11:13:00Z romain $
"""

from datasource import DataSource

class TableBase(DataSource):

    """ This is a table base object.
    You can call a SGDB and receive his answer in a formated table.
    Note that it is designed for SOAP calls.

    Example:

        class MyTable(TableBase):

            def __init__(self, database):
                TableBase.__init__(database, ['hello', 'world'])

            def __call__(self, **args):

                sort = args['sort']

                result = self._sql_query('SELECT hello, world from helloworld ORDER BY %s' % sort)

                result.addCallback(self._print_result)
                return result

            def entry_form(self, entry):

                result = (entry[0] + '_',)
                result += entry[1:]
                return result

    Usage:

        >>> table = MyTable()     # Create MyTable instance
        >>> deferred = table(**{'sort': 'hello'})    # <=> table.__call__()

        If this is a SOAP function, return deferred object, and SOAP will wait
        callback function to return it value.
        _print_result() return a dict.

    """

    def __init__(self, database, columns):
        """
            @param database [DataBase] database object
            @param columns [list] list of columns id
        """

        DataSource.__init__(self, database)
        self.columns = columns
        self.table = []
        self.entries = 0
        self.rowcount = 0
        self.count = 0

        assert isinstance(self.columns, list)

    def __iter__(self):

        return self.table.__iter__()

    def entry_form(self, entry):
        """ This is a callback for each entry received by _sql_query()
        Overload this method to process a entry.

            @param entry [tuple] This is a line.

            @return [tuple]

        WARNING: tuple is an imutable object, so you HAVE to recreate another tuple
        to modify it !
        """

        return entry

    def _print_count(self, result):
        """ Callback used by a _sql_query() when I want to receive a result of a
            COUNT(*) query.
        """

        try:
            return int(result[0][0][0])
        except TypeError:
            return 0

    def _save_count(self, result):
        """ Callback used by a _sql_query() when I want to receive a result of a
            COUNT(*) query.
        """

        self.count = result[0][0][0]

        return self


    def _print_result(self, result):
        """ Callback of _sql_query() when I receive a result from SGDB
            @param result [array of tuple]

            @return [dict] If called by SOAP, client will receive result of this function
                            ret['args'] = dict with all arguments used for this request
                            ret['filters'] = dict with all filters
                            ret['columns'] = list of column titles
                            ret['rowcount'] = integer number of lines (it doesn't consider 'start' and 'limit')
                            ret['table'] = 2D array with table resulted
        """

        self.rowcount = result[1]
        result = result[0]

        for entry in result:
            self.table += [self.entry_form(entry)]

        return self

    def _remove_column(self, name):

        try:
            self.columns.remove(name)
        except:
            pass

    def _add_column(self, name):

        self.columns += [name]

class InfoBase(TableBase):
    """ This class is herited from TableBase and has just only one specific overloaded function """


    def _print_result(self, result):
        """ Callback of _sql_query() when I receive a result from SGDB
            In this function, we will create a dict to associate key and value.
            @param result [array of tuple]

            @return [dict] dictionnary with all key/values results
        """

        rowcount = result[1]
        result = result[0]

        for entry in result:
            self.table += [self.entry_form(entry),]

        if len(self.table) > 0:
            self.info = dict()
            for key, value in zip(self.columns, self.table[0]):
                self.info[key] = value
        else:
            return None

        return self
